/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __CAMERA_H__
#define __CAMERA_H__

#include "ray.h"
#include "vector2f.h"

class Camera {
	Vector w[3];
	Vector rowinc, colinc;
	Vector cam;
	void calculate_bases(void);
	Vector point_rotate(Vector v);
public:
	Vector look_at;
	double alpha, beta;
	double dist;
	double fov;

	Camera();
	void init(void);
	void frame_init(void);
	
	Ray get_screen_ray(double screen_x, double screen_y);
	Vector row_increment(void) const;
	Vector col_increment(void) const;
	
	// camera movement
	void pan(int mouse_x, int mouse_y);
	void rotate(int mouse_x, int mouse_y);
	void dolly(int mouse_y);
	
	// rasterizing ops
	
	// Given a point in 3D space, determine where it would 
	// end up on the 2D screen
	vec2f reverse(const Vector& v);
	
	void rotate_real(double d_alpha, double d_beta);
};


extern Camera camera;

#endif //__CAMERA_H__
