/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include "config.h"

#define prop_read(prop) \
	if (!strcmp(#prop, name)) sscanf(value, "%d", &prop)

#define prop_write(prop) \
	fprintf(f, "%s=%d\n", #prop, prop)
	
#define prop_readf(prop) \
	if (!strcmp(#prop, name)) sscanf(value, "%f", &prop)

#define prop_writef(prop) \
	fprintf(f, "%s=%.5f\n", #prop, prop)


void Config::defaults(void)
{
	xres = 640;
	yres = 480;
	aspect = 4.0/3.0;
	aa = 0;
	shadows = 0;
	reflections = 0;
	procedural = 0;
	mouse_sensitivity = 0.05;
	speed = 2.5;
	bilinear = 1;
	fullscreen = 0;
	focal_blur = 0;
	blur_samples = 50;
	cel_shading = 0;
}

bool Config::init(void)
{
	defaults();
	FILE *f = fopen("config.txt", "rt");
	if (!f) return false;
	char buff[200];
	while (fgets(buff, sizeof(buff), f)) {
		int i = 0;
		while (buff[i] && buff[i] != '=') i++;
		if (buff[i] == '=') {
			buff[i] = 0;
			i++;
			char * name = buff, *value = buff + i;
			prop_read(xres);
			prop_read(yres);
			prop_readf(aspect);
			prop_read(aa);
			prop_read(shadows);
			prop_read(reflections);
			prop_read(procedural);
			prop_read(bilinear);
			prop_read(fullscreen);
			prop_read(cel_shading);
			prop_readf(mouse_sensitivity);
			prop_readf(speed);
		}
	}
	fclose(f);
	return true;
}

void Config::close(void)
{
	FILE *f = fopen("config.txt", "wt");
	prop_write(xres);
	prop_write(yres);
	prop_writef(aspect);
	prop_write(aa);
	prop_write(shadows);
	prop_write(reflections);
	prop_write(procedural);
	prop_write(bilinear);
	prop_write(fullscreen);
	prop_write(cel_shading);
	prop_writef(mouse_sensitivity);
	prop_writef(speed);
	fclose(f);
}

Config config;
