/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __DISPLAY_H__
#define __DISPLAY_H__

#include "utils.h"

struct SDL_Surface;

struct Rect {
	int x0, y0, w, h;
	Rect(){}
	Rect(int x0, int y0, int w, int h)
	{
		this->x0 = x0;
		this->y0 = y0;
		this->w = w;
		this->h = h;
	}
	bool has(int x, int y);
};

struct ClickCallback {
	virtual void process(int x, int y) = 0;
	virtual ~ClickCallback(){}
};

struct Clickable {
	Rect rect;
	ClickCallback *callback;
};

class Display {
	SDL_Surface *surface;
	unsigned *buff;
	int x, y;
public:
	bool init(void);
	void close(void);
	unsigned * buffer(void);
	void update(void);
	void clear(unsigned color = 0);
	void screenshot(void);
};

enum {
	BLEFT = 1,
	BRIGHT = 2,
	BMIDDLE = 4,
	CLICKABLE_MASK = 8
};

class GUI {
	unsigned color;
	int cx, cy;
	void xwrite(const char *what);
	int mousebuttons;
	Clickable clickables[20];
	int ccount;
public:
	GUI();
	void pause(void);
	void sleep(int ms);
	void font_color(unsigned);
	void write(const char * format, ...);
	void writexy(int x, int y, const char * format, ...);
	void movexy(int x, int y);
	void handle_inputs(bool & should_exit);
	bool pending_events(void);
	double btime(void);
	void add_clickable(Rect rect, ClickCallback* callback);
	void remove_clickable(ClickCallback *callback);
	void remove_all_clickables(void);
	bool cursor_inside(Rect rect);
};

extern Display display;
extern GUI ui;

#endif //__DISPLAY_H__
