/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <string.h>
#include "game.h"
#include "camera.h"
#include "display.h"
#include "render.h"
#include "config.h"
#include "pawn.h"
#include "bitmap.h"
#include "playerchooser.h"

Game game;

class HumanPlayer: public Player
{
public:
	int think_move(int pos[4][4], int me, int steps)
	{
		bool she = false;
		int valid[4], n = 0;
		for (int i = 0; i < 4; i++) 
			if (game.valid_move(pos[me][i], steps))
				valid[n++] = me * 4 + i;
		renderer.enable_glow(valid, n);
		game.display_board();
		struct UserPawnClick : public ClickCallback{
			volatile int answer;
			int valid[4], n;
			UserPawnClick() { answer = -1; }
			void process(int x, int y) {
				int id = renderer.get_id_at(x, y);
				for (int i = 0; i < n; i++)
					if (valid[i] == id)
						answer = id;
			}
		} upc;
		upc.n = n;
		memcpy(upc.valid, valid, sizeof(valid));
		ui.add_clickable(Rect(0,0,config.xres, config.yres), &upc);
		//
		while (upc.answer == -1) {
			if (ui.pending_events()) {
				ui.handle_inputs(she);
				game.display_board();
			}
			ui.sleep(50);
			if (she) break;
		}
		//
		if (she)
			game.exit();
		renderer.disable_glow();
		ui.remove_clickable(&upc);
		if (upc.answer == -1)
			return -1;
		return upc.answer % 4;
	}
	bool is_human(void) { return true; }
	~HumanPlayer(){}
};

class ComputerPlayer: public Player
{
public:
	int think_move(int pos[4][4], int me, int steps)
	{
		/*
		 * In order to keep things fast, the logic of the computer player
		 * is hardcoded. The following rules are obeyed, in decreasing
		 * order of importance:
		 *
		 * 1) If you can capture, do so. If you have a choice what to
		 *    capture, order by:
		 *      1.1) Number of pawns
		 *      1.2) Pawn "oldness" (old pawns are more pleasant food)
		 *
		 * 2) If an old enough pawn is threatened by other pawn behind it,
		 *    avoid by moving forward. Conversely, if moving forward
		 *    makes your pawn threatened, avoid moving it.
		 *
		 * 3) If you strike six, and there are less than three pawns in the
		 *    "battlefield", respawn one
		 *
		 * 4) If movement is not constrained, move older pawns first. This
		 *    may outweigh 3) sometimes.
		*/
		int board[56] = {0};
		// Board number legend:
		// + 1000 for every enemy pawn on the cell
		// + X for the oldness of the pawn(s) on the cell
		// + 1 if the cell is threatened
		// we ignore our pawns, they don't play in captures or threatening
		for (int plr = 0; plr < 4; plr++) if (plr != me) {
			bool has_homies = false;
			for (int i = 0; i < 4; i++) {
				if (pos[plr][i] < 0) has_homies = true;
				if (pos[plr][i] < 0 || pos[plr][i] >= 100) continue;
				int j = game.absolutize(plr, pos[plr][i]);
				board[j] += 1000 + pos[plr][i];
				j = pos[plr][i];
				for (int k = 0; k < 6; k++) {
					j = game.step_position(j, 1);
					if (j < 100)
						board[game.absolutize(plr, j)]++;
				}
			}
			// mark every respawning cell as threatened
			if (has_homies)
				board[game.absolutize(plr, 0)]++;
		}
	
		
		//
		int scores[4] = { 0 };
		for (int i = 0; i < 4; i++) {
			if (!game.valid_move(pos[me][i], steps)) continue;
			int a = pos[me][i];
			int age = a;
			int xa = game.absolutize(me, a);
			int b = game.step_position(a, steps);
			int xb = game.absolutize(me, b);
			// check for captures
			if (xb < 100) {
				if (board[xb] >= 1000) {
					// bonus for capture
					scores[i] += board[xb];
				} else {
					// penalty getting into a threatened cell
					if (age > 18 && board[xb])
						scores[i] -= 100 * board[xb];
				}
			}
			if (steps == 6) {
				// we struck a SIX...
				int in_home = 0, in_battle = 0, in_heaven = 0;
				for (int k = 0; k < 4; k++) {
					if (pos[me][k] < 0) ++in_home;
					if (pos[me][k] >= 0 && pos[me][k] < 56) ++in_battle;
					if (pos[me][k] >= 100) ++in_heaven;
				}
				if (in_battle < 3 && pos[me][i] < 0) {
					// bonus for respawn, if reasonable
					scores[i] += 125;
				}
			}
			if (xa < 100) { // if not-in-heaven
				// bonus for getting out of a threatened cell
				if (age > 18 && board[xa])
					scores[i] += 100 * board[xa];
				
				// apply rule 4)
				if (age >= 0)
					scores[i] += 3*age/2;
			}
			
			// if anyone will get to heaven directly, get a small bonus:
			if (b == 105)
				scores[i] += 20;
		}
		// see which is the best pawn to move:
		int bi = -1, smax = -1000;
		for (int i = 0; i < 4; i++) if (game.valid_move(pos[me][i], steps)) {
			if (scores[i] > smax) {
				smax = scores[i];
				bi = i;
			}
		}
		/*
		printf("Computer player %d: my pawns are at %5d, %5d, %5d and %5d\n",
		       me, pos[me][0], pos[me][1], pos[me][2], pos[me][3]);
		printf("My evaluation of scores is         %5d, %5d, %5d and %5d\n",
		       scores[0], scores[1], scores[2], scores[3]);
		printf("I've chosen to move pawn #%d\n", bi);
		*/
		return bi;
	}
	bool is_human(void) { return false; }
	~ComputerPlayer(){}
};

const Vector pos_home[4] = {
	Vector(3.9, 0, -3.9),
	Vector(5.1, 0, -3.9),
	Vector(3.9, 0, -5.1),
	Vector(5.1, 0, -5.1),
};

const Vector pos_heaven[4] = {
	Vector(1.18, 0, 0.68),
	Vector(1.18, 0,-0.68),
	Vector(0.52, 0, 0.0),
	Vector(1.18, 0, 0.0),
};
	

static Vector player_rotate(int player, Vector res)
{
	double x = res[0], z = res[2];
	switch (player) {
		case 0: return Vector(x, 0, z);
		case 1: return Vector(-z, 0, x);
		case 2: return Vector(-x, 0, -z);
		case 3: return Vector(z, 0, -x);
		default:
			return Vector(0,0,0);
	}
}

Vector Game::get_position_real(int player, int index)
{
	if (index < 0)
		return player_rotate(player, pos_home[-index-1]);
	if (index >= 105)
		return player_rotate(player, pos_heaven[index-105]);
	if (index >= 100 && index < 105)
		return player_rotate(player, Vector(6-index + 100, 0, 0));
	if (index < 14) {
		Vector res;
		if (index == 13)
			res = Vector(0, 0, 7);
		if (index < 13)
			res = Vector(1, 0, 3 + index - 8);
		if (index == 6)
			res = Vector(2, 0, 2);
		if (index < 6)
			res = Vector(7 - index, 0, 1);
		return player_rotate(player, res);
	} else {
		return get_position_real((player+1) % 4, index - 14);
	}
}

Vector Game::get_position(int player, int idx)
{
	int index = pos[player][idx];
	if (index == 105) {
		for (int j = 0; j < idx; j++) {
			if (pos[player][j] == pos[player][idx]) {
				index++;
			}
		}
	}
	return get_position_real(player, index);
}

Vector Game::get_full_position(int i, int j)
{
	Vector r = get_position(i, j);
	for (int k = 0; k < j; k++)
		if (pos[i][k] == pos[i][j] && pos[i][j] >= 0 && pos[i][j] != 105)
			r[1] += 0.6;
	return r;
}

Game::Game() {
	memset(players, 0, sizeof(players));
	game_loaded = false;
	num_human = 1;
	num_comp = 14;
}

void game_preinit(void)
{
	game.preinit();
}

struct PersistentNewGameCallback : public ClickCallback {
	void process(int x, int y)
	{
		game.on_new_game();
	}
};
static PersistentNewGameCallback persistent_new_game_callback;

void Game::preinit(void)
{
	init_newgame_bitmap();
}

void Game::setup_game(void)
{
	Rect rect;
	rect.w = newgame_bmp->width();
	rect.h = newgame_bmp->height();
	rect.y0 = 0;
	rect.x0 = config.xres - rect.w;
	ui.remove_all_clickables();
	ui.add_clickable(rect, &persistent_new_game_callback);
}

Game::~Game()
{
	for (int i = 0; i < 4; i++) {
		if (players[i]) delete players[i];
		players[i] = NULL;
	}
	if (newgame_bmp) delete newgame_bmp;
	newgame_bmp = NULL;
}

void Game::init_newgame_bitmap(void)
{
	if (newgame_bmp) return;
	newgame_bmp = new Bitmap;
	newgame_bmp->load_embedded("newgame.button.bmp");
	Rect rect = get_dice_rect();
	if (newgame_bmp->width() != rect.w) {
		newgame_bmp->rescale(rect.w, 0);
	}
	unsigned *d = newgame_bmp->data();
	for (int j = 0; j < newgame_bmp->height(); j++)
		for (int i = 0; i < newgame_bmp->width(); i++, d++)
			if (i == 0 || j == 0 || i == newgame_bmp->width()-1 || j == newgame_bmp->height() - 1)
				*d = 0xffffff;
}


void Game::init(void)
{
	who = 0;
	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < 4; j++) {
			pos[i][j] = -1-j;
		}
		if (players[i]) {
			delete players[i];
		}
		players[i] = NULL;
	}
	dice_value = 0;
	dont_save_state = false;
}

void Game::new_game(void)
{
	init();
	camera.init();
	for (int i = 0; i < 4; i++) {
		if (num_human & (1 << i)) {
			players[i] = new HumanPlayer;
		} else {
			if (num_comp & (1 << i))
				players[i] = new ComputerPlayer;
		}
	}
}

void Game::on_new_game(void)
{
	exit();
	dont_save_state = true;
}

bool select_players(void)
{
	if (!game.load_state()) {
		game.init();
		if (!choose_player_colors(game.num_human, game.num_comp))
			return false;
		game.new_game();
	}
	game.setup_game();
	return true;
}

bool play_game(void)
{
	return game.play_game();
}

bool Game::play_game(void)
{
	dont_save_state = false;
	she = false;
	while (!she) {
		ui.handle_inputs(she);
		if (she) break;
		if (players[who]) {
			display_board();
			process_player();
		}
		if (she) break;
		if (last_dice_num != 6 || !players[who]) {
			who++; who %= 4;
		}
	}
	if (dont_save_state) {
		remove("gamestate.txt");
	} else {
		save_state();
	}
	return dont_save_state;
}

void Game::display_board(void)
{
	renderer.reinit_scene();
	for (int i = 0; i < 4; i++) if (players[i]) {
		for (int j = 0; j < 4; j++) {
			Vector where = get_full_position(i, j);
			renderer.add_pawn(where, i, i * 4 + j);
		}
	}
	renderer.render();
	render_gui();
	display.update();
}

void Game::render_gui(void)
{
	if (newgame_bmp) {
		int w = newgame_bmp->width();
		int h = newgame_bmp->height();
		int x0 = config.xres - w;
		for (int i = 0; i < h; i++) {
			unsigned * src = newgame_bmp->data() + i * w;
			unsigned * dest= display.buffer() + x0 + i * config.xres;
			memcpy(dest, src, w * sizeof(unsigned));
		}
	}
	if (dice_value != 0)
		render_dice();
}

int Game::next_position(int i)
{
	if (i < 0) return 0;
	if (i < 55) return i + 1;
	if (i == 55) return 100;
	if (i < 105) return i + 1;
	return i;
}

bool Game::valid_move(int i, int x)
{
	return step_position(i, x) != i;
}

int Game::step_position(int i, int x)
{
	if (i < 0) {
		if (x == 6) return 0;
		return i;
	}
	if (i >= 100) {
		if (i + x <= 105) {
			return i + x;
		} else return i;
	}
	int j = i + x;
	if (j >= 56) {
		j = 100 + j - 56;
		if (j <= 105) return j;
		else return i;
	}
	return i + x;
}

void Game::process_player(void)
{
	//pos[who][0] = next_position(pos[who][0]);
	bool gl = game_loaded; game_loaded = false;
	int x;
	if (!gl || exitstate == 0) {
		show_dice('?');
		if (players[who]->is_human()) {
			request_dice_click();
		}
		if (she) {
			exitstate = 0;
			return;
		}
	
		x = dice_random.next();
		last_dice_num = x;
		animate_dice(x, players[who]->is_human());
	} else {
		x = last_dice_num;
		animate_dice(x, players[who]->is_human());
	}
	int possible_moves = 0;
	for (int i = 0; i < 4; i++) {
		if (valid_move(pos[who][i], x))
			++possible_moves;
	}
	if (!possible_moves) {
		if (players[who]->is_human()) {
			sleep_with_display_updates(500);
		}
		return;
	}
	int r;
	if (possible_moves > 1 && (!gl || exitstate == 2)) {
		r = players[who]->think_move(pos, who, x);
	} else {
		r = 0;
		while (!valid_move(pos[who][r], x)) r++;
	}
	if (gl && exitstate == 1) return;
	if (r != -1) {
		exitstate = 1;
		process_move(r, x);
	} else {
		exitstate = 2;
	}
}


static float dice_inside(float x, float y, char c)
{
	#undef CIRCLE
	#define CIRCLE_r(x0, y0, t0, t1, t2) \
		if (sqr(x0-x) + sqr(y0-y) < t0) { \
			float d = sqr(x0-x) + sqr(y0-y); \
			if (d < t1) \
				return 1.0f; \
			else {\
				return 1.0f-(d - t1) / t2;\
			}\
		}
	#define CIRCLE(x0, y0) CIRCLE_r(x0,y0,0.01f,0.008f,0.002f)
	#define SMALL_CIRCLE(x0, y0) CIRCLE_r(x0,y0,0.005f,0.003f,0.002f)
	if (c >= '1' && c <= '6') {
		int idx = c - '1';
		vec2f a[6][6] = {
			{ vec2f(0,0)},
			{ vec2f(.3, .3), vec2f(-.3, -.3)},
			{ vec2f(.3, .3), vec2f(0,0), vec2f(-.3, -.3)},
			{ vec2f(.3, .3), vec2f(.3,-.3), vec2f(-.3, .3), vec2f(-.3, -.3)},
			{ vec2f( .3,  .3), vec2f( .3, -.3),
			             vec2f(0,0), 
			  vec2f(-.3,  .3), vec2f(-.3, -.3)},
			{ vec2f(-.3, .3), vec2f(.3, .3),
			  vec2f(-.3, 0), vec2f(.3, 0),
			  vec2f(-.3, -.3), vec2f(.3, -.3),
			}
		};
		for (int i = 0; i <= idx; i++) {
			CIRCLE(a[idx][i].x, a[idx][i].y);
		}
	}
	if (c == '?') {
		SMALL_CIRCLE(0, 0.3);
		if (y > -0.4 && y < 0.15) {
			y = (y + 0.4f) / (0.15f + 0.4f);
			y = sqr(y);
			y *= 2*M_PI;
			float x0 = sin(y) * 2 / (1 + y*8);
			float d = fabs(x - x0);
			if (d < 0.07f) {
				if (d < 0.06f)
					return 1.0f;
				else
					return 1.0f - (d - 0.06f) / 0.01f;
			}
		}
	}
	return 0.0f;
	#undef CIRCLE
}

Rect Game::get_dice_rect(void)
{
	int h = config.xres / 8;
	int x0 = config.xres - h - 1;
	int y0 = h / 2;
	return Rect(x0, y0, h, h);
}

void Game::render_dice(void)
{
	if (!(dice_value == '?' || dice_value == ' ' || (dice_value >= '1' && dice_value<='6')))
		return;
	unsigned dice_color = mul(player_colors[who], 0.5f);
	unsigned * b = display.buffer();
	Rect rect = get_dice_rect();
	int h = rect.w;
	int x0 = rect.x0;
	int y0 = rect.y0;
	for (int j = 0; j <= h; j++) {
		for (int i = 0; i <= h; i++) {
			unsigned &res = b[(y0 + j) * config.xres + x0 + i];
			if (i == 0 || i == h || j == 0 || j == h) {
				res = 0xffffff;
			} else {
				float x = i / (float) h - 0.5f;
				float y = j / (float) h - 0.5f;
				x *= 1.05; y *= 1.05;
				double dist = x*x*x*x + y*y*y*y;
				if (dist > 0.0625f) {
					res = 0x333344;
				} else {
					float coeff;
					unsigned wc;
					if (dist > 0.0500f) {
						coeff = (dist - 0.050f) / 0.0125f;
						wc = 0x333344;
					} else {
						coeff = dice_inside(x, y, dice_value);
						wc = 0xffffff;
					}
					res = mul(wc, coeff) + mul(dice_color, 1-coeff);
				}
			}
		}
	}
	display.update();
}

void Game::show_dice(char value)
{
	dice_value = value;
	render_dice();
}

void Game::animate_dice(int new_value, bool is_human)
{
	show_dice(' ');
	sleep_with_display_updates(333);
	show_dice('0' + new_value);
	if (!is_human) sleep_with_display_updates(500);
}

void Game::process_move(int r, int x)
{
	AnimationEntry a[25];
	int n = 0;
	int i = pos[who][r];
	if (i < 0) x = 1;
	pos[who][r] = 99; // invalid, but temporary
	int time = 0;
	while (x) {
		x--;
		int next = next_position(i);
		a[n].id = who * 4 + r;
		a[n].start_time = time;
		a[n].end_time = time + 1;
		a[n].start_pos = get_position_real(who, i);
		a[n].start_pos += Vector(0, 0.6, 0) * get_num_pawns(absolutize(who, i));
		
		if (next == 105) {
			for (int l = 0; l < 4; l++)
				if (pos[who][l] == 105)
					next++;
		}
		a[n].end_pos = get_position_real(who, next);
		a[n].end_pos += Vector(0, 0.6, 0) * get_num_pawns(absolutize(who, next));
		n++;
		time ++;
		if (next > 105) next = 105;
		i = next;
	}
	pos[who][r] = i;
	int m = 0;
	int idx[4];
	for (int p = 0; p < 4; p++) if (p != who) {
		for (int j = 0; j < 4; j++) {
			if (absolutize(p, pos[p][j]) == absolutize(who, pos[who][r])) {
				idx[m++] = j;
			}
		}
		if (!m) continue;
		a[n].id = who * 4 + r;
		a[n].start_time = time;
		a[n].end_time = time+2;
		a[n].start_pos = a[n].end_pos = get_position_real(who, i);
		a[n].start_pos += Vector(0, 0.6, 0) * get_num_pawns(absolutize(who, i));
		n++;
		for (int k = 0; k < m; k++) {
			int r = idx[k];
			a[n].id = p * 4 + r;
			a[n].start_time = time;
			a[n].end_time = time + 2;
			a[n].start_pos = get_position_real(who, i) + Vector(0, 0.6, 0) * k;
			a[n].end_pos = free_home(p, r);
			n++;
		}
		break;
	}
	class ShowDiceCallback : public Callback {
		void process(void)
		{
			game.render_gui();
		}
	} cb;
	renderer.animate(a, n, she, cb);
}

int Game::absolutize(int player, int i)
{
	if (i < 0) return i - player * 100;
	if (i >= 100) return i + player * 100;
	return (i + player * 14) % 56;
}

int Game::get_num_pawns(int absolutepos)
{
	int c = 0;
	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < 4; j++) {
			if ((pos[i][j] >= 0 && pos[i][j] < 56) || (pos[i][j] >= 100 && pos[i][j] < 105))
				if (absolutize(i, pos[i][j]) == absolutepos)
					c++;
		}
	}
	return c;
}

Vector Game::free_home(int player, int i)
{
	int c = 0;
	int u[4] = {0,0,0,0};
	for (int j = 0; j < 4; j++) {
		if (pos[player][j] < 0) {
			u[-pos[player][j]-1] = true;
		}
	}
	while (u[c]) c++;
	Vector ret = get_position_real(player, -1-c);
	pos[player][i] = -1-c;
	return ret;
}

void Game::request_dice_click(void)
{
	
	struct ClickWaiter: public ClickCallback {
		volatile bool clicked;
		void process(int x, int y)
		{
			clicked = true;
		}
		
	} cw;
	cw.clicked = false;
	ui.add_clickable(get_dice_rect(), &cw);
	while (!cw.clicked) {
		if (she) break;
		if (ui.pending_events()) {
			ui.handle_inputs(she);
			display_board();
		}
		ui.sleep(50);
	}
	ui.remove_clickable(&cw);
}

void Game::exit(void)
{
	she = true;
}

void Game::sleep_with_display_updates(unsigned ms)
{
	while (!she && ms > 0) {
		if (ui.pending_events()) {
			ui.handle_inputs(she);
			display_board();
		}
		unsigned sleep_time = ms > 50 ? 50 : ms;
		ms -= sleep_time;
		ui.sleep(sleep_time);
	}
}

void benchmark(void)
{
	game.init();
	game.new_game();
	game.show_dice('?');
	double t0 = ui.btime();
	int i = 0;
	bool counting = false;
	while (i < 100) {
		if (!counting && ui.btime() - t0 > 2.5) { // allow heat-up
			counting = true;
			t0 = ui.btime();
		}
		camera.rotate_real(0.07, 0);
		game.display_board();
		if (counting)
			++i;
	}
	t0 = ui.btime() - t0;
	printf("%.2lf FPS\n", (double) i / t0);
}

bool command_args_allow_playing_the_game(int argc, char **argv)
{
	if (argc == 2) {
		if (!strcmp(argv[1], "--bench")) {
			benchmark();
			return false;
		}
		if (!strcmp(argv[1], "--spline")) {
			test_splines();
			return false;
		}
	}
	return true;
}


/*
	Game save file structure
	
	<player1>
	<player2>
	<player3>
	<player4>
	<who> <exitstate> <last_dice_num>
	<camera>
	
	<player> ::= <playertype> <pawn1> <pawn2> <pawn3> <pawn4>
	<playertype> ::= 0 - NULL, 1 - human, 2 - comp
	<who> - 0|1|2|3
	<camera> ::= <cam.alpha> <cam.beta> <cam.dist> <cam.look_at>
*/
void Game::save_state(void)
{
	FILE *f = fopen("gamestate.txt", "wt");
	for (int i = 0; i < 4; i++) {
		int playertype = 0;
		if (players[i])
			playertype = players[i]->is_human()? 1 : 2;
		fprintf(f, "%d %d %d %d %d\n",
			playertype, pos[i][0], pos[i][1], pos[i][2], pos[i][3]);
	}
	fprintf(f, "%d %d %d\n", who, exitstate, last_dice_num);
	fprintf(f, "%.5lf %.5lf %.5lf %.5lf %.5lf %.5lf\n",
		camera.alpha,
		camera.beta,
		camera.dist,
		camera.look_at.v[0],
		camera.look_at.v[1],
		camera.look_at.v[2]);
	fclose(f);
}

bool Game::load_state(void)
{
	FILE *f = fopen("gamestate.txt", "rt");
	if (!f) return false;
	bool good = true;
	bool hasplayers = false;
	for (int i = 0; i < 4; i++) {
		int playertype = 0;
		good = good && (5 == fscanf(f, "%d %d %d %d %d\n",
			&playertype, &pos[i][0], &pos[i][1], &pos[i][2], &pos[i][3]));
		if (playertype) {
			hasplayers = true;
			if (playertype == 1)
				players[i] = new HumanPlayer;
			else
				players[i] = new ComputerPlayer;
		}
	}
	if (!hasplayers) {
		good = false;
		for (int i = 0; i < 4; i++) {
			if (players[i])
				delete players[i];
			players[i] = NULL;
		}
	}
	good = good && (3 == fscanf(f, "%d %d %d\n", &who, &exitstate, &last_dice_num));
	good = good && (6 == fscanf(f, "%lf %lf %lf %lf %lf %lf\n",
		&camera.alpha,
		&camera.beta,
		&camera.dist,
		&camera.look_at.v[0],
		&camera.look_at.v[1],
		&camera.look_at.v[2]));
	fclose(f);
	game_loaded = good;
	return good;
}

