/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __GAME_H__
#define __GAME_H__

#include "display.h"
#include "vector3.h"

enum { HUMAN = 1, COMPUTER = 2 };

class Player;
class Bitmap;

class Game {
	bool she;
	char dice_value;
	int last_dice_num;
	int exitstate;
	bool game_loaded;
	Bitmap *newgame_bmp;
	void init_newgame_bitmap(void);
	bool dont_save_state;

	
public:
	Game();
	~Game();
	int num_human, num_comp;
	Player* players[4];
	int pos[4][4];
	int who;
//
	void init(void);
	void preinit(void);
	void new_game(void);
	void on_new_game(void);
	void setup_game(void);
	bool play_game(void);
	void display_board(void);
	void process_player(void);
	void sleep_with_display_updates(unsigned ms);
	
	int next_position(int i);
	bool valid_move(int i, int x);
	int step_position(int i, int x);
	
	Vector get_position_real(int i, int index);
	Vector get_position(int i, int j);
	Vector get_full_position(int i, int j);
	
	void render_dice(void);
	void render_gui(void);
	void show_dice(char value);
	void animate_dice(int new_value, bool is_human);
	
	void process_move(int r, int x);
	
	int absolutize(int player, int i);
	int get_num_pawns(int absolutepos);
	Vector free_home(int player, int i);
	void request_dice_click(void);
	Rect get_dice_rect(void);
	void exit(void);
	
	bool load_state(void);
	void save_state(void);
};

class Player {
public:
	/**
	 * Represents the thinking process when choosing to move one of several
	 * pawns.
	 *
	 * @param pos - the positions of pawns of all players
	 * @param me  - the current player #
	 * @param steps - how many steps the pawn has to move forward
	 * @returns the index of the pawn the player chose to move
	 * NOTE: the player will have at least two legal choices
	 */
	virtual int think_move(int pos[4][4], int me, int steps) = 0;
	/// returns true if the player (should be) human (for UI, etc)
	virtual bool is_human(void) = 0;
	virtual ~Player() {};
};

extern Game game;

bool select_players(void);
bool play_game(void);
void game_preinit(void);
bool command_args_allow_playing_the_game(int argc, char **argv);

#endif // __GAME_H__
