/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>
#include "matrix.h"

Matrix rotation(int axis, double angle)
{
	double sina = sin(angle);
	double cosa = cos(angle);
	Matrix m(1);
	switch (axis) {
		case AXIS_X:
			m.m[1][1] = cosa;
			m.m[1][2] = sina;
			m.m[2][1] = -sina;
			m.m[2][2] = cosa;
			break;
		case AXIS_Y:
			m.m[0][0] = cosa;
			m.m[0][2] = -sina;
			m.m[2][0] = sina;
			m.m[2][2] = cosa;
			break;
		case AXIS_Z:
			m.m[0][0] = cosa;
			m.m[0][1] = sina;
			m.m[1][0] = -sina;
			m.m[1][1] = cosa;
			break;
		default:
			/* nothing */
			break;
	}
	return m;
}

double Matrix::det(void) const
{
	return 
	 + m[0][0] * m[1][1] * m[2][2]
	 - m[0][0] * m[1][2] * m[2][1]
	 - m[0][1] * m[1][0] * m[2][2]
	 + m[0][1] * m[1][2] * m[2][0]
	 + m[0][2] * m[1][0] * m[2][1]
	 - m[0][2] * m[1][1] * m[2][0];
}

double Matrix::adj(int ii, int jj) const
{
	int rows[2], rc = 0, cols[2], cc = 0;
	for (int i = 0; i < 3; i++)
		if (i != ii) rows[rc++] = i;
	for (int j = 0; j < 3; j++)
		if (j != jj) cols[cc++] = j;
	double t = m[rows[0]][cols[0]] * m[rows[1]][cols[1]] - m[rows[1]][cols[0]] * m[rows[0]][cols[1]];
	if ((ii + jj) % 2) t = -t;
	return t; 
}

Matrix inverse(const Matrix& m)
{
	double D = m.det();
	if (fabs(D) < 1e-12) return m;
	double rD = 1.0 / D;
	Matrix result;
	for (int i = 0; i < 3; i++)
		for (int j = 0; j < 3; j++)
			result.m[i][j] = rD * m.adj(j, i);
	return result;
}
