/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdlib.h>
#include "pawn.h"
#include "ray.h"
#include "display.h"
#include "config.h"
#include "utils.h"

struct NaturalSpline {
	double *coeffs;
	double *x;
	int n;
	
	void xNaturalSpline(double x[], double y[], double coeff_points[], int nn) // nn = number of points - 1
	{
		this->n = nn;
		this->x = (double*) malloc(sizeof(double) * (n+1));
		this->coeffs = (double*) malloc(sizeof(double) * 4 * n);
		//
		for (int i = 0; i <= n; i++)
			this->x[i] = x[i];
		//
		for (int i = 1; i < n-1; i++) {
			for (int j = 0; j < 4; j++) {
				coeffs[i*4+j] = coeff_points[i*4-4+j];
			}
		}
	}
	
	NaturalSpline(double x[], double y[], double coeff_points[], int nn)
	{
		xNaturalSpline(x, y, coeff_points, nn);
	}
	
	~NaturalSpline()
	{
		free(x);
		free(coeffs);
	}
	
	double eval(double X)
	{
		if (X < x[1]) X = x[1];
		if (X > x[n-1]) X = x[n-1];
		int i = 2;
		while (x[i] < X) i++;
		i--;
		double p = X;
		return (((coeffs[4*i]*p) + coeffs[4*i+1])*p + coeffs[4*i+2]) * p + coeffs[4*i+3];
	}
	
	double eval(int i, double X)
	{
		return (((coeffs[4*i]*X) + coeffs[4*i+1])*X + coeffs[4*i+2]) * X + coeffs[4*i+3];
	}
	
	double eval_deriv(double X)
	{
		if (X < x[1]) X = x[1];
		if (X > x[n-1]) X = x[n-1];
		int i = 2;
		while (x[i] < X) i++;
		i--;
		// return 3*A*X*X + 2*A*X + B
		return (((3 * coeffs[4*i] * X) + 2 * coeffs[4*i+1]) * X) + coeffs[4*i+2];
	}
};




static double x_points[] = { -0.05, 0.00, 0.10, 0.80, 0.90, 1.00, 1.10 };
static double y_points[] = { -0.05, 0.00, 0.07, 0.24, 0.38, 0.40, 0.48 };
static double coeff_points[] = {
	   -5.5999999999999996447,    1.0499999999999998224,
	    0.0000000000000000000,    0.0000000000000000000,
	    0.1580174927113703331,   -0.1104664723032070944,
	    0.0593527696793003171,   -0.0000886297376093278,
	  -94.2000000000281261237,  242.3200000000719285254,
	 -206.6620000000612833446,   58.5328000000173886974,
	   69.5999999999996248334, -199.3999999999989256594,
	  190.3999999999987551291,  -60.4399999999998982503};
static NaturalSpline spline(x_points, y_points, coeff_points, sizeof(x_points)/sizeof(x_points[0]) - 1);

void test_splines(void)
{
	unsigned *b = display.buffer();
	//for (int i = 20; i < config.yres - 20; i++) {
	for (double z = 0; z <= 1.0; z += 0.0001) {
		unsigned color = 0xffffff;
		for (unsigned j = 0; j < sizeof(x_points)/sizeof(x_points[0]); j++)
			if (fabs(z-x_points[j]) < 0.005) color = 0xff0000;
		int i = 20 + (int) (z * (config.yres - 40));
		//double z = (i - 20) / (config.yres - 40.0);
		double off = spline.eval(z);
		off = sqrt(off);
		int ioff = (int) (off * (config.yres - 40.0));
		int idx = i * config.xres + config.xres/2;
		b[idx-ioff] = color;
		b[idx+ioff] = color;
	}
	display.update();
	ui.pause();
}

static inline double cubeeval(double a, double b, double c, double d, double x)
{
	return ((((a * x) + b) * x + c) * x + d);
}

const double r3 = 0.333333333333333333, r9 = 0.11111111111111111111111,
	r27 = 0.037037037037037037, r04 = 2.5, r54 = 0.0185185185185185185;
	


#define CBRT2 1.2599210498948731648             /* 2^(1/3) */
#define SQR_CBRT2 1.5874010519681994748         /* 2^(2/3) */

static const float factor[5] =
{
  1.0f / SQR_CBRT2,
  1.0f / CBRT2,
  1.0f,
  CBRT2,
  SQR_CBRT2
};


// Copied from GNU Libc
static float
cube_root (float x)
{
  float xm, ym, u, t2;
  int xe;

  /* Reduce X.  XM now is an range 1.0 to 0.5.  */
  xm = frexpf (fabsf (x), &xe);

  /* If X is not finite or is null return it (with raising exceptions
     if necessary.
     Note: *Our* version of `frexp' sets XE to zero if the argument is
     Inf or NaN.  This is not portable but faster.  */
  if (xe == 0 && fpclassify (x) <= FP_ZERO)
    return x + x;

  u = (0.492659620528969547f + (0.697570460207922770f
                               - 0.191502161678719066f * xm) * xm);

  t2 = u * u * u;

  ym = u * (t2 + 2.0f * xm) / (2.0f * t2 + xm) * factor[2 + xe % 3];

  return ldexpf (x > 0.0f ? ym : -ym, xe / 3);
}

// Copied from the GNU scientific library
static int cubic_solve_gsl(double a, double b, double c, double *X)
{
  double q = (a * a - 3 * b);
  double r = (2 * a * a * a - 9 * a * b + 27 * c);

  double Q = q * r9;
  double R = r * r54;

  double Q3 = Q * Q * Q;
  double R2 = R * R;

  double CR2 = 729 * r * r;
  double CQ3 = 2916 * q * q * q;
  double ar3 = a*r3;

  if (R == 0 && Q == 0)
    {
      X[0] = - ar3 ;
      return 1 ;
    }
  else if (CR2 == CQ3) 
    {
      /* this test is actually R2 == Q3, written in a form suitable
         for exact computation with integers */

      /* Due to finite precision some double roots may be missed, and
         considered to be a pair of complex roots z = x +/- epsilon i
         close to the real axis. */

      double sqrtQ = sqrt (Q);

      if (R > 0)
        {
          X[0] = -2 * sqrtQ  - ar3;
          X[1] = sqrtQ - ar3;
        }
      else
        {
          X[0] = - sqrtQ  - ar3;
          X[1] = 2 * sqrtQ - ar3;
        }
      return 2 ;
    }
  else if (CR2 < CQ3) /* equivalent to R2 < Q3 */
    {
      double sqrtQ = sqrt (Q);
      double sqrtQ3 = sqrtQ * sqrtQ * sqrtQ;
      double theta = acos (R / sqrtQ3);
      double norm = -2 * sqrtQ;
      X[0] = norm * cos (theta * r3) - ar3;
      X[1] = norm * cos ((theta + 2.0 * M_PI) * r3) - ar3;
      X[2] = norm * cos ((theta - 2.0 * M_PI) * r3) - ar3;
      return 3;
    }
  else
    {
      double sgnR = (R >= 0 ? 1 : -1);
      double A = -sgnR * cube_root (fabs (R) + sqrt (R2 - Q3));
      double B = Q / A ;
      X[0] = A + B - ar3;
      return 1;
    }
}


static int cubic_solve(double A, double B, double C, double D, double x[])
{
	if (A == 0) {
		double a = B;
		double b = C;
		double c = D;
		double D = b*b - 4 * a * c;
		if (D < 0) return 0;
		D = sqrt(D);
		double ra = 0.5 / a;
		x[0] = (-b - D) * ra;
		x[1] = (-b + D) * ra;
		return 2;
	}
	if (A != 1.0) {
		double rA = 1.0 /A;
		B *= rA;
		C *= rA;
		D *= rA;
	}
	return cubic_solve_gsl(B, C, D, x);
}

	
double Pawn::intersect(const Ray & xray, IContext &ctx)
{
	Ray ray = xray;
	if (skew_amount != 0) {
		ray.c -= pos;
		ray.c = iuvw * ray.c;
		ray.c += pos;
		ray.dir = iuvw * ray.dir;
	}
	// d^2 = sqr(pos.x - ray.c.x - p * ray.dir.x) + (.y...)
	double R2 = radius * radius;
	double hx = pos[0] - ray.c[0];
	double hy = pos[2] - ray.c[2];
	double A = sqr(ray.dir[0]) + sqr(ray.dir[2]);
	double B = -2 * (hx * ray.dir[0] + hy * ray.dir[2]);
	double C = sqr(hx) + sqr(hy);
	double r2A = 0.5 / A;
	
	// distance from pawn's center is A*p^2 + B * p + C. The closest point is -b/2a
	double X = -B * r2A;
	double cd = A*X*X+B*X+C;
	if (cd > R2 || (X < 0 && C > R2)) return 1e99; // it doesn't cross the bounding cyl - bail out
	
	double scaling = radius * r04;
	double scale2 = sqr(scaling);
	
	C -= R2;
	double D = B * B - 4 * A * C;
	if (D < 0) return 1e99;
	D = sqrt(D);
	double x1 = ( -B - D) * r2A;
	double x2 = ( -B + D) * r2A;
	Vector p1 = ray.c + ray.dir * x1;
	Vector p2 = ray.c + ray.dir * x2;
	if ((p1[1] < pos[1]  && p2[1] < pos[1]) || (p1[1] > pos[1] + scaling && p2[1] > pos[1] + scaling)) {
		return 1e99;
	}
	C += R2;
	//
	/*
	ctx.ip = p1;
	ctx.id = 1;
	ctx.object = this;
	*((Vector*)&ctx.extra) = p1 - pos;
	return ctx.ip.distto(ray.c);
	*/
	//
	double Y0 = pos[1];
	double Y1 = pos[1] + scaling;
	double mint, maxt, dt;
	bool extremal_case = false;
	double rd1;
	if (fabs(ray.dir[1]) < 5*1e-3) {
		mint = -1e99;
		maxt = 1e99;
		dt = 1e99;
		extremal_case = true;
	} else {
		rd1 = 1.0 / ray.dir[1];
		mint = (Y1 - ray.c[1]) * rd1;
		maxt = (Y0 - ray.c[1]) * rd1;
		dt = maxt - mint;
	}
	double rmi = min(mint, maxt);
	double rma = max(mint, maxt);
	
	
	
	double min_p = 1e99;
	double deriv = 0;
	if (extremal_case) {
		double x = (ray.c[1] - Y1) / (Y0-Y1);
		double off = spline.eval(x)*scale2;
		C -= off;
		D = B * B - 4 * A * C;
		if (D < 0) return 1e99;
		D = sqrt(D);
		min_p = (- B - D) / (2 * A);
		if (off < 1e-9)
			deriv = 1e99;
		else
			deriv = 0.5 * spline.eval_deriv(x) * scale2 / sqrt(off);
	} else {
		for (int i = 1; i < spline.n - 1; i++) {
			double U, V, W, T;
			U = -spline.coeffs[i*4+0]*scale2;
			V = -spline.coeffs[i*4+1]*scale2;
			W = -spline.coeffs[i*4+2]*scale2;
			T = -spline.coeffs[i*4+3]*scale2;
			U += 0;
			V += A * dt * dt;
			W += 2 * A * dt * mint + B * dt;
			T += A * mint * mint + B * mint + C;
			
			double sols[3];
			int sc = cubic_solve(U, V, W, T, sols);
			for (int j = 0; j < sc; j++) {
				if (sols[j] < spline.x[i] || sols[j] > spline.x[i+1]) continue;
				double p = (maxt-mint)*sols[j] + mint;
				if (p < min_p && p > 0 && p >= rmi && p <= rma) {
					min_p = p;
					double off = spline.eval(i, sols[j])*scale2;
					if (off < 1e-9)
						deriv = 1e99;
					else
						deriv = 0.5 * spline.eval_deriv(sols[j]) * scale2 / sqrt(off);
				}
			}
		}
	}
	
	if (min_p < 1e99) {
		ctx.ip = xray.c + xray.dir * min_p;
		ctx.id = id;
		ctx.object = this;
		Vector & normal = *((Vector*) &ctx.extra);
		if (skew_amount != 0) {
			normal = ray.c + ray.dir * min_p - pos;
		} else {
			normal = ctx.ip - pos;
		}
		normal[1] = 0;
		normal.norm();
		double co = sqrt(1.0 / (deriv * deriv + 1.0));
		double si = sqrt(1.0 - co * co);
		normal = normal * co + Vector(0, si, 0);
		if (skew_amount != 0) normal = uvw * normal;
		return ctx.ip.distto(xray.c);
	}
	return 1e99;
}

unsigned Pawn::shade(const Ray & ray, IContext & ctx, const Light & light)
{
	// phong = ambient + diffuse * L * N + specular * (R * V) ^ n
	unsigned color = this->color;
	Vector &I = ctx.ip;
	Vector L = light.pos - I; L.norm();
	Vector &N = *((Vector*) &ctx.extra);
	Vector R = L*-1.0; R += N * 2.0; R.norm();
	Vector V = ray.c - I;
	float diffuse = (L * N);
	//float total = 0.35f; //ambient
	float total;
	float yy = I[1];
	if (yy < 0) yy = 0; if (yy > 1) yy = 1;
	total = 0.65f * cosf(-1.1f + 2.0f * sqrtf(yy));
	if (!(config.shadows && renderer.in_shadow(ctx.ip, this))) {
		if (diffuse >= 0) {
			float specular = (R * V) / V.length();
			if (specular < 0.0f) specular = 0.0f;
			specular *= specular;specular *= specular; specular *= specular; specular *= specular;
			total += specular * 5.8f + diffuse * 0.8f;
		}
	}
	if (config.reflections) {
		total *= 0.75f;
		Vector T = I - ray.c;
		T.norm();
		T += N * 2.0;
		unsigned refl = renderer.raytrace(Ray(I, T), this);
		if (refl) {
			unsigned rc, gc, bc, rr, gr, br;
			bc = color & 0xff; br = refl & 0xff; color >>= 8; refl >>= 8;
			gc = color & 0xff; gr = refl & 0xff; color >>= 8; refl >>= 8;
			rc = color & 0xff; rr = refl & 0xff;
			bc = min((unsigned) (bc * total) + (unsigned)(br * total), 255u);
			gc = min((unsigned) (gc * total) + (unsigned)(gr * total), 255u);
			rc = min((unsigned) (rc * total) + (unsigned)(rr * total), 255u);
			color = bc | (gc << 8) | (rc << 16);
		} else {
			color = mul(color, total);
		}
		return color;
	}
	return mul(color, total);
}


Pawn::~Pawn(){}

BBox Pawn::get_bbox(void)
{
	double r = radius;
	if (skew_amount != 0)
		r *= 1.25;
	return BBox(pos + Vector(-r, 0, -r), pos + Vector(r, r / 0.4, r));
}

int Pawn::get_id(void)
{
	return id;
}

void Pawn::move(Vector start_pos, Vector end_pos, double x)
{
	//pos = new_pos;
	Vector v0 = start_pos;
	Vector v1 = end_pos;
	Vector vd = v1 - v0;
	double scale = vd.length();
	Vector res = v0 + vd * x + Vector(0, sin(x*M_PI) * scale * 0.5, 0);
	pos = res;
	movement = vd;
	skew_amount = sin(x * M_PI)*0.5;
	//
	// prepare the matrices
	Vector r = Vector(0, 1, 0) ^ movement;
	if (r.length() < 1e-9) {
		skew_amount = 0;
		return;
	}
	r.norm();
	double angle = atan2(r[2], r[0]);
	uvw = rotation(AXIS_Y, angle) * rotation(AXIS_X, skew_amount) * rotation(AXIS_Y, -angle);
	iuvw = inverse(uvw);
}
