/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <string.h>
#include "display.h"
#include "bitmap.h"
#include "config.h"

Rect std_rects[14] = {
	Rect( 410, 127, 52, 18 ), Rect( 477, 127, 73, 18 ), Rect( 563, 127, 55, 18 ),
	Rect( 410, 191, 52, 18 ), Rect( 477, 191, 73, 18 ), Rect( 563, 191, 55, 18 ),
	Rect( 410, 255, 52, 18 ), Rect( 477, 255, 73, 18 ), Rect( 563, 255, 55, 18 ),
	Rect( 410, 319, 52, 18 ), Rect( 477, 319, 73, 18 ), Rect( 563, 319, 55, 18 ),
	Rect( 158, 391, 70, 30 ), Rect( 417, 391, 63, 30 ),
};

class Chooser {
public:
	virtual void notify(int id) = 0;
	virtual ~Chooser(){}
};
struct PCCallback : public ClickCallback {
	Chooser *pc;
	int id;
	PCCallback(Chooser *chooser, int idd): pc(chooser), id(idd) {}
	void process(int x, int y) {
		pc->notify(id);
	}
};

class PlayerChooser : public Chooser
{
	int &hh;
	int &cc;
	int human, comp;
	Rect r[14];
	int x, y;
	int bx, by, bw, bh;
	PCCallback * callbacks[14];
	double lastinside[14], firstinside[14];
	volatile bool done, result;
	
public:
	PlayerChooser(int &h, int &c) : hh(h), cc(c) { human = h, comp = c;}
	~PlayerChooser ()
	{
		for (int i = 0; i < 14; i++)
			delete callbacks[i];
	}
	bool doit(void)
	{
		done = false;
		x = config.xres;
		y = config.yres;
		Bitmap bmp;
		if (!bmp.load_embedded("choice.bmp")) {
			return false;
		}
		int ox = bmp.width(), oy = bmp.height();
		if (bmp.width() > x)
			bmp.rescale(x, 0);
		if (bmp.height() > y)
			bmp.rescale(0, y);
		float s = min((float)bmp.width() / ox, (float)bmp.height() / oy);
		if (s != 1.0f) {
			for (int i = 0; i < 14; i++) {
				r[i].x0 = (int) (s*std_rects[i].x0+0.5f);
				r[i].y0 = (int) (s*std_rects[i].y0+0.5f);
				r[i].w = (int) (s*std_rects[i].w+0.5f);
				r[i].h = (int) (s*std_rects[i].h+0.5f);
			}
		} else {
			memcpy(r, std_rects, sizeof(r));
		}
		bw = bmp.width();
		bh = bmp.height();
		bx = (x - bw) / 2;
		by = (y - bh) / 2;
		for (int i = 0; i < 14; i++) {
			r[i].x0 += bx;
			r[i].y0 += by;
		}
		ui.remove_all_clickables();
		for (int i = 0; i < 14; i++) {
			callbacks[i] = new PCCallback(this, i);
			ui.add_clickable(r[i], callbacks[i]);
		}
		Bitmap screen(x, y);
		unsigned *sd = screen.data();
		unsigned *bd = bmp.data();
		for (int j = 0; j < bh; j++)
			for (int i = 0; i < bw; i++)
				sd[(j+by) * x + i + bx] = bd[j*bw+i];
		for (int i = 0; i < 14; i++)
			lastinside[i] = firstinside[i] = -1;
		
		bool she = false;
		result = true;
		while (!done) {
			render(screen);
			ui.handle_inputs(she);
			double xtime = ui.btime();
			for (int i = 0; i < 14; i++) {
				if (ui.cursor_inside(r[i])) {
					lastinside[i] = xtime;
					if (firstinside[i] == -1)
						firstinside[i] = xtime;
				} else
					firstinside[i] = -1;
			}
		}
		hh = human;
		cc = comp;
		return result;
	}
	void notify(int id)
	{
		if (id >= 12) {
			if (id == 13) 
				result = false;
			done = true;
		} else {
			int plr = id / 3;
			int what = id % 3;
			int nh = human, nc = comp;
			int mask = 1 << plr;
			switch (what) {
				case 0:
					nh &= ~mask;
					nc &= ~mask;
					break;
				case 1:
					nh |= mask;
					nc &= ~mask;
					break;
				case 2:
					nh &= ~mask;
					nc |= mask;
					break;
				default:
					break;
			}
			if (nh || nc) {
				human = nh;
				comp = nc;
			}
		}
	}
	void render(Bitmap &screen)
	{
		unsigned *src = screen.data();
		unsigned *dest = display.buffer();
		double t = ui.btime();
		for (int j = 0; j < y; j++) {
			for (int i = 0; i < x; i++, src++, dest++) {
				float f = get_intensity(i, j, t);
				if (f != 1.0) {
					*dest = mul(*src, f);
				} else {
					*dest = *src;
				}
			}
		}
		display.update();
	}
	float get_intensity(int xi, int yi, double t)
	{
		int idx = 13;
		while (idx >= 0) {
			if (r[idx].has(xi, yi)) break;
			idx--;
		}
		if (idx < 0) return 1.0f;
		if (is_full(idx)) return 2.65f;
		if (firstinside[idx] == -1) {
			double el = (t - lastinside[idx])*2;
			if (el > 1) return 1.0f;
			else return 1.0f + 0.6f*cos(el * M_PI/2);
		} else {
			double el = (t - firstinside[idx])*2;
			if (el > 1) return 1.6f;
			else return 1.0f + 0.6f*cosf((1-el) * M_PI/2);
		}
	}
	bool is_full(int idx)
	{
		if (idx >= 12) return false;
		int plr = idx / 3;
		int what = idx % 3;
		int msk = 1 << plr;
		if (what == 0 && ((human & msk) == 0) && ((comp & msk) == 0)) return true;
		if (what == 1 && ((human & msk) != 0) && ((comp & msk) == 0)) return true;
		if (what == 2 && ((human & msk) == 0) && ((comp & msk) != 0)) return true;
		return false;
	}
};


bool choose_player_colors(int &human, int &comp)
{
	PlayerChooser pc(human, comp);
	return pc.doit();
}
