/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __UTILS_H__
#define __UTILS_H__

#include <stdlib.h>
#include "vector3.h"

#undef sqr
template <typename T>
static inline T sqr(T x) { return x*x; }

template <typename T>
static inline T min(T a, T b) { return a < b ? a : b; }

template <typename T>
static inline T max(T a, T b) { return a > b ? a : b; }

static inline unsigned mul(unsigned c, float f)
{
	float b = c & 0xff; c>>=8;
	float g = c & 0xff; c>>=8;
	float r = c & 0xff;
	b *= f;
	g *= f;
	r *= f;
	if (b > 255.0f) b = 255.0f;
	if (g > 255.0f) g = 255.0f;
	if (r > 255.0f) r = 255.0f;
	return (((int)r) << 16) + (((int)g) << 8) + ((int)b);
}

struct Random {
	int dummy;
	Random();
	void seed(unsigned seed);
	int next(void);
	double drand(void);
};

class Callback {
public:
	virtual void process(void) = 0;
	virtual ~Callback() {}
};

extern Random dice_random;

#undef abs
template <typename T>
static inline T abs(T x) {
	return x < 0 ? -x : x;
}

#define SSSTORAGE_SIZE (1024*1024)
extern char static_sort_storage[SSSTORAGE_SIZE];

	
template <typename T>
void _mergesort_imp(T a[], int l, int r, T t[])
{
	/* Stage 1 : check for bottom */
	if (r <= l) return;
	if (r == l + 1) {
		if (a[r] < a[l]) {
			t[0] = a[l];
			a[l] = a[r];
			a[r] = t[0];
		}
		return;
	}
	
	/* Stage 2 : recurse down */
	int m = (l + r) / 2;
	_mergesort_imp(a, l, m, t);
	_mergesort_imp(a, m + 1, r, t);
	
	/* Stage 3 : merge */
	int size = (r - l + 1);
	int i = l, j = m + 1, k = 0;
	while (i <= m && j <= r) {
		if (a[j] < a[i])
			t[k++] = a[j++];
		else
			t[k++] = a[i++];
	}
	if (i <= m) for (; i <= m; i++) t[k++] = a[i];
	if (j <= r) for (; j <= r; j++) t[k++] = a[j];
	for (k = 0; k < size; k++)
		a[l + k] = t[k];
}
	
/** Generic sorting function
 *  THREAD SAFETY: No (uses static storage)
 *  STABLE: Yes
 *  @param a - the array of values to sort
 *  @param n - the number of elements in the array
 * NOTE: sort() requires a operator &lt; to be defined
*/
template <typename T>
void sort(T a[], int n)
{
	// if no sorting is needed
	if (n < 2) return;
	
	// try using the static storage...
	T* storage = (T*) static_sort_storage;
	bool static_storage = true;
	
	if (sizeof(T) * n > SSSTORAGE_SIZE) {
		// .. but if it is not enough, use a dynamic one
		static_storage = false;
		storage = (T*) malloc(sizeof(T)*n);
	}
	
	// perform the actual sortng
	_mergesort_imp(a, 0, n - 1, storage);
	
	// non static storage must be freed
	if (!static_storage)
		free(storage);
}

/**
 * perpendicular - returns a vector, orthogonal to @p v
 * @note - since there are infinitely many such vectors, the returned one is
 *         constrained to have two of its components being 0 and 1, respectively
 * @note - another (distinct) orthogonal vector might be obtained using
 *         Vector ortho2 = v ^ perpendicular(v);
 * 
 * @param v - the vector, which perpendicular is being asked
 * @returns null if v is null
 *          a orthogonal vector to v, otherwise
 */ 
Vector perpendicular(const Vector &v);


#endif
